"""
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import numpy as np
import pytest

from fixtures.fields import iter_grids
from pde import FieldCollection, ScalarField, Tensor2Field, UnitGrid, VectorField
from pde.fields.base import FieldBase
from pde.tools.misc import skipUnlessModule


@pytest.mark.parametrize("grid", iter_grids())
def test_shapes_nfields(grid, rng):
    """test single component field"""
    for num in [1, 3]:
        fields = [ScalarField.random_uniform(grid, rng=rng) for _ in range(num)]
        field = FieldCollection(fields)
        data_shape = (num,) + grid.shape
        np.testing.assert_equal(field.data.shape, data_shape)
        for pf_single in field:
            np.testing.assert_equal(pf_single.data.shape, grid.shape)

        field_c = field.copy()
        np.testing.assert_allclose(field.data, field_c.data)
        assert field.grid == field_c.grid


def test_collections(rng):
    """test field collections"""
    grid = UnitGrid([3, 4])
    sf = ScalarField.random_uniform(grid, label="sf", rng=rng)
    vf = VectorField.random_uniform(grid, label="vf", rng=rng)
    tf = Tensor2Field.random_uniform(grid, label="tf", rng=rng)
    fields = FieldCollection([sf, vf, tf])
    assert fields.data.shape == (7, 3, 4)
    assert isinstance(str(fields), str)

    fields2 = FieldCollection({"s": sf, "v": vf, "t": tf}, copy_fields=True)
    assert fields == fields2
    assert fields2.labels == ["s", "v", "t"]
    assert not np.shares_memory(fields[0].data, fields2[0].data)

    copy = fields[:2]
    assert isinstance(copy, FieldCollection)
    assert len(copy) == 2
    np.testing.assert_equal(copy[0].data, sf.data)
    assert not np.may_share_memory(copy[0].data, sf.data)
    np.testing.assert_equal(copy[1].data, vf.data)
    assert not np.may_share_memory(copy[1].data, vf.data)

    copy = fields[2:]
    assert isinstance(copy, FieldCollection)
    assert len(copy) == 1
    np.testing.assert_equal(copy[0].data, tf.data)
    assert not np.may_share_memory(copy[0].data, tf.data)

    fields.data[:] = 0
    np.testing.assert_allclose(sf.data, 0)
    np.testing.assert_allclose(vf.data, 0)
    np.testing.assert_allclose(tf.data, 0)

    assert fields[0] is fields["sf"]
    assert fields[1] is fields["vf"]
    assert fields[2] is fields["tf"]
    with pytest.raises(KeyError):
        fields["42"]

    sf.data = 1
    vf.data = 1
    tf.data = 1
    np.testing.assert_allclose(fields.data, 1)
    assert all(np.allclose(i, 12) for i in fields.integrals)
    assert all(np.allclose(i, 1) for i in fields.averages)
    assert np.allclose(fields.magnitudes, np.sqrt([1, 2, 4]))

    assert sf.data.shape == (3, 4)
    assert vf.data.shape == (2, 3, 4)
    assert tf.data.shape == (2, 2, 3, 4)

    c2 = FieldBase.from_state(fields.attributes, data=fields.data)
    assert c2 == fields
    assert c2.grid is grid

    attrs = FieldCollection.unserialize_attributes(fields.attributes_serialized)
    c2 = FieldCollection.from_state(attrs, data=fields.data)
    assert c2 == fields
    assert c2.grid is not grid

    fields["sf"] = 2.0
    np.testing.assert_allclose(sf.data, 2)
    with pytest.raises(KeyError):
        fields["42"] = 0

    fields.plot(subplot_args=[{}, {"scale": 1}, {"colorbar": False}])


def test_collections_copy():
    """test copying data of collections"""
    grid = UnitGrid([2, 2])
    sf = ScalarField(grid, 0)
    vf = VectorField(grid, 1)
    fc = FieldCollection([sf, vf])

    data = np.r_[np.zeros(4), np.ones(8)]
    np.testing.assert_allclose(fc.data.flat, data)

    fc2 = fc.copy()
    assert fc.data is not fc2.data
    assert fc[0].data is not fc2[0].data
    assert fc[1].data is not fc2[1].data

    sf.data = 1
    np.testing.assert_allclose(fc.data.flat, np.ones(12))
    np.testing.assert_allclose(fc2.data.flat, data)

    # special case
    fc = FieldCollection([sf, sf])
    fc[0] = 2
    np.testing.assert_allclose(fc[0].data, 2)
    np.testing.assert_allclose(fc[1].data, 1)


def test_collections_append():
    """test append data to collections"""
    grid = UnitGrid([2, 2])
    sf = ScalarField(grid, 0)
    vf = VectorField(grid, 1, label="vector")
    c1 = FieldCollection([sf], labels=["scalar"])

    c2 = c1.append(vf)
    assert len(c2) == 2 and len(c1) == 1
    data = np.r_[np.zeros(4), np.ones(8)]
    np.testing.assert_allclose(c2.data.flat, data)

    assert c1.data is not c2.data
    assert c1[0].data is not c2[0].data
    assert vf.data is not c2[1].data
    assert list(c2.labels) == ["scalar", "vector"]

    c3 = c1.append(c1, label="new")
    assert len(c3) == 2 and len(c1) == 1
    np.testing.assert_allclose(c3.data.flat, np.zeros(8))

    assert c1.data is not c3.data
    assert c1[0].data is not c3[0].data
    assert vf.data is not c3[1].data
    assert c3.label == "new"

    c4 = c1.append(c1, vf)
    assert len(c4) == 3 and len(c1) == 1
    data = np.r_[np.zeros(8), np.ones(8)]
    np.testing.assert_allclose(c4.data.flat, data)


def test_collections_operators():
    """test field collections"""
    grid = UnitGrid([3, 4])
    sf = ScalarField(grid, 1)
    vf = VectorField(grid, 1)
    fields = FieldCollection([sf, vf])

    fields += fields
    np.testing.assert_allclose(fields.data, 2)
    np.testing.assert_allclose(sf.data, 2)
    np.testing.assert_allclose(vf.data, 2)

    fields = fields - 1
    np.testing.assert_allclose(fields.data, 1)

    fields = fields + fields
    np.testing.assert_allclose(fields.data, 2)

    fields *= 2
    np.testing.assert_allclose(fields.data, 4)


def test_smoothing_collection(rng):
    """test smoothing of a FieldCollection"""
    grid = UnitGrid([3, 4], periodic=[True, False])
    sf = ScalarField.random_uniform(grid, rng=rng)
    vf = VectorField.random_uniform(grid, rng=rng)
    tf = Tensor2Field.random_uniform(grid, rng=rng)
    fields = FieldCollection([sf, vf, tf])
    sgm = 0.5 + rng.random()

    out = fields.smooth(sigma=sgm)
    for i in range(3):
        np.testing.assert_allclose(out[i].data, fields[i].smooth(sgm).data)

    out.data = 0
    fields.smooth(sigma=sgm, out=out)
    for i in range(3):
        np.testing.assert_allclose(out[i].data, fields[i].smooth(sgm).data)


def test_scalar_random_uniform():
    """test creating collections using scalar_random_uniform"""
    grid = UnitGrid([3, 4], periodic=[True, False])
    fc = FieldCollection.scalar_random_uniform(2, grid, label="c", labels=["a", "b"])
    assert fc.label == "c"
    assert fc[0].label == "a"
    assert fc[1].label == "b"
    assert fc[0].grid is grid
    assert fc[1].grid is grid
    assert not np.allclose(fc[0].data, fc[1].data)


def test_from_scalar_expressions():
    """test creating field collections from scalar expressions"""
    grid = UnitGrid([3])
    expressions = ["x**2", "1"]
    fc = FieldCollection.from_scalar_expressions(
        grid, expressions=expressions, label="c", labels=["a", "b"]
    )
    assert fc.label == "c"
    assert fc[0].label == "a"
    assert fc[1].label == "b"
    assert fc[0].grid is grid
    assert fc[1].grid is grid
    np.testing.assert_allclose(fc[0].data, (np.arange(3) + 0.5) ** 2)
    np.testing.assert_allclose(fc[1].data, 1)


@skipUnlessModule("napari")
@pytest.mark.interactive
def test_interactive_collection_plotting(rng):
    """test the interactive plotting"""
    grid = UnitGrid([3, 3])
    sf = ScalarField.random_uniform(grid, 0.1, 0.9, rng=rng)
    vf = VectorField.random_uniform(grid, 0.1, 0.9, rng=rng)
    field = FieldCollection([sf, vf])
    field.plot_interactive(viewer_args={"show": False, "close": True})


def test_field_labels():
    """test the FieldCollection.labels property"""
    grid = UnitGrid([5])
    s1 = ScalarField(grid, label="s1")
    s2 = ScalarField(grid)
    fc = FieldCollection([s1, s2])

    assert fc.labels == ["s1", None]
    assert len(fc.labels) == 2
    assert fc.labels[0] == "s1"
    assert fc.labels.index("s1") == 0
    assert fc.labels.index(None) == 1
    with pytest.raises(ValueError):
        fc.labels.index("a")

    fc.labels = ["a", "b"]
    assert fc.labels == ["a", "b"]
    fc.labels[0] = "c"
    assert fc.labels == ["c", "b"]
    assert str(fc.labels) == str(["c", "b"])
    assert repr(fc.labels) == repr(["c", "b"])

    assert fc.labels[0:1] == ["c"]
    assert fc.labels[:] == ["c", "b"]
    fc.labels[0:1] = "d"
    assert fc.labels == ["d", "b"]

    fc.labels[:] = "a"
    assert fc.labels == ["a", "a"]

    labels = fc.labels[:]
    labels[0] = "e"
    assert fc.labels == ["a", "a"]

    fc = FieldCollection([s1, s2], labels=[None, "b"])
    assert fc.labels == [None, "b"]
    fc = FieldCollection([s1, s2], labels=["a", "b"])
    assert fc.labels == ["a", "b"]

    with pytest.raises(TypeError):
        fc.labels = [1, "b"]
    with pytest.raises(TypeError):
        fc.labels[0] = 1


def test_collection_1_field():
    """test field collections with only one field"""
    grid = UnitGrid([3])
    s1 = ScalarField(grid, label="a")
    fc = FieldCollection([s1])
    assert fc.labels == ["a"]

    fc.plot()


def test_collection_plotting():
    """test simple plotting of various fields on various grids"""
    grid = UnitGrid([5])
    s1 = ScalarField(grid, label="s1")
    s2 = ScalarField(grid)
    fc = FieldCollection([s1, s2])

    # test setting different figure sizes
    fc.plot(figsize="default")
    fc.plot(figsize="auto")
    fc.plot(figsize=(3, 3))

    # test different arrangements
    fc.plot(arrangement="horizontal")
    fc.plot(arrangement="vertical")


def test_from_data(rng):
    """test the `from_data` method"""
    grid = UnitGrid([3, 5])
    s = ScalarField.random_uniform(grid, label="s1", rng=rng)
    v = VectorField.random_uniform(grid, label="v2", rng=rng)
    f1 = FieldCollection([s, v])

    f2 = FieldCollection.from_data(
        [ScalarField, VectorField],
        grid,
        data=f1.data,
        with_ghost_cells=False,
        labels=["a", "b"],
    )
    assert f2.labels == ["a", "b"]
    np.testing.assert_allclose(f1.data, f2.data)

    f3 = FieldCollection.from_data(
        [ScalarField, VectorField],
        grid,
        data=f1._data_full,
        with_ghost_cells=True,
        labels=["c", "d"],
    )
    assert f3.labels == ["c", "d"]
    np.testing.assert_allclose(f1.data, f2.data)


def test_collection_apply(rng):
    """test the `apply` method"""
    grid = UnitGrid([3, 5])
    s = ScalarField(grid, 2, label="s1")
    v = VectorField.random_uniform(grid, label="v2", rng=rng)
    f1 = FieldCollection([s, v])

    np.testing.assert_allclose(f1.apply("s1 * v2").data, v.data * 2)



r"""
Kuramoto-Sivashinsky - Using custom class
=========================================

This example implements a scalar PDE using a custom class. We here consider the
`Kuramoto–Sivashinsky equation
<https://en.wikipedia.org/wiki/Kuramoto–Sivashinsky_equation>`_, which for instance 
describes the dynamics of flame fronts:

.. math::
    \partial_t u = -\frac12 |\nabla u|^2 - \nabla^2 u - \nabla^4 u
"""

from pde import PDEBase, ScalarField, UnitGrid


class KuramotoSivashinskyPDE(PDEBase):
    """Implementation of the normalized Kuramoto–Sivashinsky equation"""

    def evolution_rate(self, state, t=0):
        """implement the python version of the evolution equation"""
        state_lap = state.laplace(bc="auto_periodic_neumann")
        state_lap2 = state_lap.laplace(bc="auto_periodic_neumann")
        state_grad = state.gradient(bc="auto_periodic_neumann")
        return -state_grad.to_scalar("squared_sum") / 2 - state_lap - state_lap2


grid = UnitGrid([32, 32])  # generate grid
state = ScalarField.random_uniform(grid)  # generate initial condition

eq = KuramotoSivashinskyPDE()  # define the pde
result = eq.solve(state, t_range=10, dt=0.01)
result.plot()



r"""
Brusselator - Using custom class
================================

This example implements the `Brusselator 
<https://en.wikipedia.org/wiki/Brusselator>`_ with spatial coupling,

.. math::

    \partial_t u &= D_0 \nabla^2 u + a - (1 + b) u + v u^2 \\
    \partial_t v &= D_1 \nabla^2 v + b u - v u^2

Here, :math:`D_0` and :math:`D_1` are the respective diffusivity and the
parameters :math:`a` and :math:`b` are related to reaction rates.

Note that the PDE can also be implemented using the :class:`~pde.pdes.pde.PDE`
class; see :doc:`the example <pde_brusselator_expression>`. However, that implementation
is less flexible and might be more difficult to extend later.
"""

import numba as nb
import numpy as np

from pde import FieldCollection, PDEBase, PlotTracker, ScalarField, UnitGrid


class BrusselatorPDE(PDEBase):
    """Brusselator with diffusive mobility"""

    def __init__(self, a=1, b=3, diffusivity=[1, 0.1], bc="auto_periodic_neumann"):
        super().__init__()
        self.a = a
        self.b = b
        self.diffusivity = diffusivity  # spatial mobility
        self.bc = bc  # boundary condition

    def get_initial_state(self, grid):
        """prepare a useful initial state"""
        u = ScalarField(grid, self.a, label="Field $u$")
        v = self.b / self.a + 0.1 * ScalarField.random_normal(grid, label="Field $v$")
        return FieldCollection([u, v])

    def evolution_rate(self, state, t=0):
        """pure python implementation of the PDE"""
        u, v = state
        rhs = state.copy()
        d0, d1 = self.diffusivity
        rhs[0] = d0 * u.laplace(self.bc) + self.a - (self.b + 1) * u + u**2 * v
        rhs[1] = d1 * v.laplace(self.bc) + self.b * u - u**2 * v
        return rhs

    def _make_pde_rhs_numba(self, state):
        """nunmba-compiled implementation of the PDE"""
        d0, d1 = self.diffusivity
        a, b = self.a, self.b
        laplace = state.grid.make_operator("laplace", bc=self.bc)

        @nb.jit
        def pde_rhs(state_data, t):
            u = state_data[0]
            v = state_data[1]

            rate = np.empty_like(state_data)
            rate[0] = d0 * laplace(u) + a - (1 + b) * u + v * u**2
            rate[1] = d1 * laplace(v) + b * u - v * u**2
            return rate

        return pde_rhs


# initialize state
grid = UnitGrid([64, 64])
eq = BrusselatorPDE(diffusivity=[1, 0.1])
state = eq.get_initial_state(grid)

# simulate the pde
tracker = PlotTracker(interval=1, plot_args={"vmin": 0, "vmax": 5})
sol = eq.solve(state, t_range=20, dt=1e-3, tracker=tracker)



"""
Special module for defining an interactive tracker that uses napari to display fields

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import logging
import multiprocessing as mp
import platform
import queue
import signal
import time
from typing import Any, Dict, Optional

from ..fields.base import FieldBase
from ..tools.docstrings import fill_in_docstring
from ..tools.plotting import napari_add_layers
from .base import InfoDict, TrackerBase
from .interrupts import IntervalData


def napari_process(
    data_channel: mp.Queue,
    initial_data: Dict[str, Dict[str, Any]],
    t_initial: Optional[float] = None,
    viewer_args: Optional[Dict[str, Any]] = None,
):
    """:mod:`multiprocessing.Process` running `napari <https://napari.org>`__

    Args:
        data_channel (:class:`multiprocessing.Queue`):
            queue instance to receive data to view
        initial_data (dict):
            Initial data to be shown by napari. The layers are named according to
            the keys in the dictionary. The associated value needs to be a tuple,
            where the first item is a string indicating the type of the layer and
            the second carries the associated data
        t_initial (float):
            Initial time
        viewer_args (dict):
            Additional arguments passed to the napari viewer
    """
    logger = logging.getLogger(__name__ + ".napari_process")

    try:
        import napari
        from napari.qt import thread_worker
    except ModuleNotFoundError:
        logger.error(
            "The `napari` python module could not be found. This module needs to be "
            "installed to use the interactive tracker."
        )
        return

    logger.info("Start napari process")

    # ignore keyboard interrupts in this process
    signal.signal(signal.SIGINT, signal.SIG_IGN)

    if viewer_args is None:
        viewer_args = {}

    # create and initialize the viewer
    viewer = napari.Viewer(**viewer_args)
    napari_add_layers(viewer, initial_data)

    # add time if given
    if t_initial is not None:
        from qtpy.QtWidgets import QLabel

        label = QLabel()
        label.setText(f"Time: {t_initial}")
        viewer.window.add_dock_widget(label)
    else:
        label = None

    def check_signal(msg: Optional[str]):
        """helper function that processes messages by the listener thread"""
        if msg is None:
            return  # do nothing
        elif msg == "close":
            viewer.close()
        else:
            raise RuntimeError(f"Unknown message from listener: {msg}")

    @thread_worker(connect={"yielded": check_signal})
    def update_listener():
        """helper thread that listens to the data_channel"""
        logger.info("Start napari thread to receive data")

        # infinite loop waiting for events in the queue
        while True:
            # get all items from the queue and display the last update
            update_data = None  # nothing to update yet
            while True:
                time.sleep(0.02)  # read queue with 50 fps
                try:
                    action, data = data_channel.get(block=False)
                except queue.Empty:
                    break

                if action == "close":
                    logger.info("Forced closing of napari...")
                    yield "close"  # signal to napari process to shut down
                    break
                elif action == "update":
                    update_data = data
                    # continue running until the queue is empty
                else:
                    logger.warning(f"Unexpected action: {action}")

            # update napari view when there is data
            if update_data is not None:
                logger.debug(f"Update napari layer...")
                layer_data, t = update_data
                if label is not None:
                    label.setText(f"Time: {t}")
                for name, layer_data in layer_data.items():
                    viewer.layers[name].data = layer_data["data"]

            yield

    # start worker thread that listens to the data_channel
    update_listener()

    # start napari
    napari.run()

    logger.info("Shutting down napari process")


class NapariViewer:
    """allows viewing and updating data in a separate napari process"""

    def __init__(self, state: FieldBase, t_initial: Optional[float] = None):
        """
        Args:
            state (:class:`pde.fields.base.FieldBase`): The initial state to be shown
            t_initial (float): The initial time. If `None`, no time will be shown.
        """
        self._logger = logging.getLogger(__name__)

        # pick a suitable multiprocessing
        if platform.system() == "Darwin":
            context: mp.context.BaseContext = mp.get_context("spawn")
        else:
            context = mp.get_context()

        # create process that runs napari
        self.data_channel = context.Queue()
        initial_data = state._get_napari_data()
        viewer_args = {
            "axis_labels": state.grid.axes,
            "ndisplay": 3 if state.grid.dim >= 3 else 2,
        }
        args = (self.data_channel, initial_data, t_initial, viewer_args)

        self.proc = context.Process(target=napari_process, args=args)  # type: ignore

        # start the process in the background
        try:
            self.proc.start()
        except RuntimeError:
            print()
            print("=" * 80)
            print(
                "It looks as if the main program did not use the multiprocessing "
                "safe-guard, which is necessary on some platforms. Please protect the "
                "main code of your program in the following way:"
            )
            print("")
            print("    if __name__ == '__main__':")
            print("        code ...")
            print("")
            print("The interactive Napari viewer could not be launched.")
            print("=" * 80)
            print()
            self._logger.exception("Could not launch napari process")

    def update(self, state: FieldBase, t: float):
        """update the state in the napari viewer

        Args:
            state (:class:`pde.fields.base.FieldBase`): The new state
            t (float): Current time
        """
        if self.proc.is_alive():
            try:
                data = (state._get_napari_data(), t)
                self.data_channel.put(("update", data), block=False)
            except queue.Full:
                pass  # could not write data
        else:
            try:
                self.data_channel.get(block=False)
            except queue.Empty:
                pass

    def close(self, force: bool = True):
        """closes the napari process

        Args:
            force (bool):
                Whether to force closing of the napari program. If this is `False`, this
                method blocks until the user closes napari manually.
        """
        if self.proc.is_alive() and force:
            # signal to napari process that it should be closed
            try:
                self.data_channel.put(("close", None))
            except RuntimeError:
                pass

        self.data_channel.close()
        self.data_channel.join_thread()

        if self.proc.is_alive():
            self.proc.join()


class InteractivePlotTracker(TrackerBase):
    """Tracker showing the state interactively in napari

    Note:
        The interactive tracker uses the python :mod:`multiprocessing` module to run
        `napari <http://napari.org/>`__ externally. The multiprocessing module
        has limitations on some platforms, which requires some care when writing your
        own programs. In particular, the main method needs to be safe-guarded so that
        the main module can be imported again after spawning a new process. An
        established pattern that works is to introduce a function `main` in your code,
        which you call using the following pattern

        .. code-block:: python

            def main():
                # here goes your main code

            if __name__ == "__main__":
                main()

        The last two lines ensure that the `main` function is only called when the
        module is run initially and not again when it is re-imported.

    """

    name = "interactive"

    @fill_in_docstring
    def __init__(
        self,
        interval: IntervalData = "0:01",
        close: bool = True,
        show_time: bool = False,
    ):
        """
        Args:
            interval:
                {ARG_TRACKER_INTERVAL}
            close (bool):
                Flag indicating whether the napari window is closed automatically at the
                end of the simulation. If `False`, the tracker blocks when `finalize` is
                called until the user closes napari manually.
            show_time (bool):
                Whether to indicate the time
        """
        # initialize the tracker
        super().__init__(interval=interval)
        self.close = close
        self.show_time = show_time

    def initialize(self, state: FieldBase, info: Optional[InfoDict] = None) -> float:
        """initialize the tracker with information about the simulation

        Args:
            state (:class:`~pde.fields.FieldBase`):
                An example of the data that will be analyzed by the tracker
            info (dict):
                Extra information from the simulation

        Returns:
            float: The first time the tracker needs to handle data
        """
        if self.show_time:
            t_initial = 0 if info is None else info.get("t_start", 0)
        else:
            t_initial = None

        self._viewer = NapariViewer(state, t_initial=t_initial)
        return super().initialize(state, info=info)

    def handle(self, state: FieldBase, t: float) -> None:
        """handle data supplied to this tracker

        Args:
            state (:class:`~pde.fields.FieldBase`):
                The current state of the simulation
            t (float):
                The associated time
        """
        self._viewer.update(state, t)

    def finalize(self, info: Optional[InfoDict] = None) -> None:
        """finalize the tracker, supplying additional information

        Args:
            info (dict):
                Extra information from the simulation
        """
        self._viewer.close(force=self.close)
